use crate::prelude::*;
use anyhow::{Context as _, Result, anyhow, ensure};
use std::{io::Write as _, process::Command};

pub(crate) fn main(nickname: Option<&str>) -> Result<()> {
    if let Some(nickname) = nickname {
        let path = app_common::try_server_config_path().context("can't find config dir")?;
        let text = format!(
            "[server]
nickname = {nickname}
"
        );

        write_new(&path, &text).context("Didn't install LookAround server.ini")?;
    }

    setup_autostart()?;

    Ok(())
}

#[cfg(not(target_os = "linux"))]
fn setup_autostart() -> Result<()> {
    anyhow::bail!("LookAround autostart is only implemented for systemd + GNU/Linux");
}

#[cfg(target_os = "linux")]
fn setup_autostart() -> Result<()> {
    let dirs = directories::BaseDirs::new().context("Could not find home dir")?;
    let home = dirs.config_local_dir();
    let path = home.join("systemd").join("user").join("lookaround.service");

    let exe_path =
        Utf8PathBuf::from_path_buf(std::env::current_exe().context("Can't get current_exe")?)
            .map_err(|_| anyhow!("current_exe isn't valid UTF-8"))?;
    let text = format!(
        "[Unit]
Description=LookAround

[Service]
ExecStart={exe_path} server
Restart=always

[Install]
WantedBy=default.target
"
    );

    write_new(&path, &text).context("Didn't install systemd service unit")?;

    ensure!(
        Command::new("systemctl")
            .args(["--user", "start", "lookaround"])
            .status()?
            .success(),
        "starting service failed"
    );
    ensure!(
        Command::new("systemctl")
            .args(["--user", "enable", "lookaround"])
            .status()?
            .success(),
        "enabling service failed"
    );

    Ok(())
}

fn write_new(path: &Path, content: &str) -> Result<()> {
    use std::fs;
    fs::create_dir_all(
        path.parent()
            .context("Impossible, path should always have a parent")?,
    )?;
    match fs::File::create_new(path) {
        Err(err) if err.kind() == std::io::ErrorKind::AlreadyExists => {
            eprintln!("File already exists, won't overwrite: `{path:?}`")
        }
        Err(err) => Err(err)?,
        Ok(mut f) => f.write_all(content.as_bytes())?,
    }
    Ok(())
}
